/*
   BEPUphysicsDemos Project

      Copyright (C) 2008 Bepu Entertainment LLC

      This demonstration software, excluding the BEPUphysics.dll and BEPUphysics.xml (which are handled independently by their accompanying license),
      is provided 'as-is', without any express or implied warranty.  In no event will the authors be held liable for any damages
      arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
    Ross Nordby
    norbo@bepu-games.com
    Harrison Nordby
    squashwell@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using System.Diagnostics;
using Microsoft.Xna.Framework.Input;
using BEPUphysics;
using BEPUphysics.Entities;

namespace BEPUphysicsDemos
{
    /// <summary>
    /// Represents the state of a character, such as standing or crouching.
    /// </summary>
    public class CharacterControllerState
    {
        /// <summary>
        /// Entity representing the shape of the body for this state.
        /// </summary>
        public Entity shape;
        /// <summary>
        /// Offset from the position of the character controller for the body entity in local space.
        /// </summary>
        public Vector3 shapeOffset;
        /// <summary>
        /// Detector entity at the feet of the body.  Should contain all rays used for support.
        /// Collects entities which are underfoot and usable for support.
        /// </summary>
        public Entity feetShape;
        /// <summary>
        /// Offset for the feet shape from the position of the character controller in local space.
        /// </summary>
        public Vector3 feetOffset;

        /// <summary>
        /// Represents support rays which allow the character body to move around and step over obstacles.
        /// Positions and directions are in local space.
        /// The rays stop after a distance equal to the length of their direction vector.
        /// </summary>
        public List<Ray> rays;
        /// <summary>
        /// Margin to 'fatten' support rays by.  Thicker rays help reduce the number of raycasts needed for good support detection, 
        /// but because ray intersections lie upon the thin, nonwidened ray, excessive size can decrease accuracy of ray tests and 
        /// harm control quality.
        /// </summary>
        public float margin;

     
        /// <summary>
        /// Constructs a state for a character controller.
        /// </summary>
        /// <param name="bodyShape">Physically simulated entity representing the body of the character.</param>
        /// <param name="bodyOffset">Offset from the character controller's position to the body shape's position in local space.</param>
        /// <param name="supportRays">Support rays for the character controller, in local space.</param> 
        /// <param name="supportMargin">Radius to 'fatten' support rays by.  Thicker rays help reduce the number of raycasts needed for good support detection, 
        /// but because ray intersections lie upon the thin, nonwidened ray, excessive size can decrease accuracy of ray tests and harm control quality.</param>
        public CharacterControllerState(Entity bodyShape, Vector3 bodyOffset, List<Ray> supportRays, float supportMargin)
        {
            shape = bodyShape;
            shapeOffset = bodyOffset;

            rays = supportRays;
            
            //Create a bounding box around the support rays.
            margin = supportMargin;
            Vector3 min = new Vector3(float.PositiveInfinity, float.PositiveInfinity, float.PositiveInfinity);
            Vector3 max = new Vector3(float.NegativeInfinity, float.NegativeInfinity, float.NegativeInfinity);
            foreach (Ray ray in rays)
            {
                min = Vector3.Min(ray.Position, min);
                min = Vector3.Min(ray.Position + ray.Direction, min);
                max = Vector3.Max(ray.Position, max);
                max = Vector3.Max(ray.Position + ray.Direction, max);
            }
            max += new Vector3(margin, margin, margin);
            min -= new Vector3(margin, margin, margin);
            //Create a detector entity using the bounding box as a basis.  The detector collects support candidates, which are then tested with a raycast.
            feetShape = new Box(Vector3.Zero, max.X - min.X, max.Y - min.Y, max.Z - min.Z, 1);
            feetOffset = (min + max) / 2;
            feetShape.isTangible = false;
            feetShape.isAffectedByGravity = false;




        }
    }

    /// <summary>
    /// Character controller based on a physically simulated entity.
    /// </summary>
    public class DynamicCharacterController:Updateable
    {
        /// <summary>
        /// Position of the character controller in the environment.
        /// </summary>
        public Vector3 position;
        /// <summary>
        /// Listing of added states.
        /// </summary>
        public Dictionary<string, CharacterControllerState> states = new Dictionary<string, CharacterControllerState>();
        /// <summary>
        /// The current state of the character controller.
        /// </summary>
        public CharacterControllerState currentState;
        /// <summary>
        /// Maximum speed that the character can attain in air through its own means.
        /// Input accleration that would otherwise exceed this value is capped.
        /// This does not affect falling or being pushed by external forces.
        /// </summary>
        public float maximumAirSpeed;
        /// <summary>
        /// Fraction of input movement to apply while in the air.
        /// </summary>
        public float airControl;
        /// <summary>
        /// Maximum speed that the character can attain while in contact with the ground through its own means.
        /// Input accleration that would otherwise exceed this value is capped.
        /// This does not affect being pushed by external forces.
        /// </summary>
        public float maximumRunSpeed;
        /// <summary>
        /// Coefficient of friction to use in slowing the character down.
        /// </summary>
        public float friction;
        /// <summary>
        /// Maximum push applied over the support rays to correct the position of the character.
        /// </summary>
        public float maximumCorrectiveImpulse;
        /// <summary>
        /// Factor of the position error to correct.  
        /// High values can cause occasional bounce, while low values can make the character feel floaty.
        /// </summary>
        public float positionCorrectionFactor;
        /// <summary>
        /// Margin in which not to apply position correction.  Helps prevent jittery motion.
        /// </summary>
        public float positionCorrectionMargin;
        /// <summary>
        /// Rotation of the character controller.
        /// </summary>
        protected Quaternion orientationQuaternion = Quaternion.Identity;
        /// <summary>
        /// Rotation of the character controller in matrix form.
        /// </summary>
        protected Matrix rotationMatrix = Matrix.Identity;
        /// <summary>
        /// List of entities which are providing the support for the character controller.
        /// </summary>
        public List<Entity> supportEntities = new List<Entity>();
        /// <summary>
        /// List of locations where a ray hit a support entity.
        /// </summary>
        public List<Vector3> supportLocations = new List<Vector3>();
        /// <summary>
        /// List of normals at locations where a ray hit a support entity.
        /// </summary>
        public List<Vector3> supportNormals = new List<Vector3>();
        /// <summary>
        /// List of distances, in units of the ray directions' lengths, of locations where a ray hit a support entity.
        /// </summary>
        public List<float> supportTois = new List<float>();
        /// <summary>
        /// List of penetrations detected by the support rays.
        /// </summary>
        public List<float> supportDepths = new List<float>();
        /// <summary>
        /// Additional 'friction' to help the character stop.
        /// Is only applied when the character was not ordered to move in the previous frame.
        /// </summary>
        public float extraSlowDownFriction;
        /// <summary>
        /// Amount to multiply the environment's friction coefficients by in interactions with the character controller.
        /// </summary>
        public float surfaceFrictionMultiplier;

        /// <summary>
        /// Whether or not the character controller was moved in the previous frame.
        /// </summary>
        public bool wasMoved = false;

        /// <summary>
        /// Whether or not this controller is activated.  When deactivated, no updates are performed.
        /// </summary>
        protected bool isActive = false;



        /// <summary>
        /// Gets and sets the rotation of the character.
        /// </summary>
        public Quaternion orientation
        {
            get
            {
                return orientationQuaternion;
            }
            set
            {
                rotationMatrix = Matrix.CreateFromQuaternion(value);
                orientationQuaternion = value;
                currentState.shape.moveTo(position);
                currentState.feetShape.moveTo(position);
                currentState.shape.orientationQuaternion = value;
                currentState.feetShape.orientationQuaternion = value;
                currentState.shape.move(Vector3.Transform(currentState.shapeOffset, rotationMatrix));
                currentState.feetShape.move(Vector3.Transform(currentState.feetOffset, rotationMatrix));
            }
        }
        /// <summary>
        /// Gets the orientation of the character controller in matrix form.
        /// </summary>
        public Matrix rotation
        {
            get
            {
                return rotationMatrix;
            }
        }

        /// <summary>
        /// Constructs a dynamic character controller.
        /// Initialize its data with the setData method.
        /// </summary>
        public DynamicCharacterController()
        {
        }

        /// <summary>
        /// Sets up the information for the character to use.
        /// </summary>
        /// <param name="pos">Position of the character controller in the environment.</param>
        /// <param name="maxAirSpeed"> Maximum speed that the character can attain in air through its own means.
        /// Input accleration that would otherwise exceed this value is capped.
        /// This does not affect falling or being pushed by external forces.</param>
        /// <param name="airControlFactor">Fraction of input movement to apply while in the air.</param>
        /// <param name="maxRunSpeed">Maximum speed that the character can attain while in contact with the ground through its own means.
        /// Input accleration that would otherwise exceed this value is capped.
        /// This does not affect being pushed by external forces.</param>
        /// <param name="frictionCoefficient">Coefficient of friction to use in slowing the character down.</param>
        /// <param name="extraSlowDown">Additional 'friction' to help the character stop.
        /// Is only applied when the character was not ordered to move in the previous frame.</param>
        /// <param name="maxCorrectiveImpulse">Maximum push applied over the support rays to correct the position of the character.</param>
        /// <param name="positionCorrectionStrength">Factor of the position error to correct.  
        /// High values provide a strong pushback on the character's feet (but can cause bounce), while low values feel 'floaty.'</param>
        /// <param name="positionCorrectionSlop">Margin in which not to apply position correction.  Helps prevent jittery motion.</param>
        /// <param name="surfaceFrictionFactor">Amount to multiply the environment's friction coefficients by in interactions with the character controller.</param>
        public void setData(Vector3 pos, float maxAirSpeed, float airControlFactor, float maxRunSpeed, float frictionCoefficient, float extraSlowDown,
            float maxCorrectiveImpulse, float positionCorrectionStrength, float positionCorrectionSlop, float surfaceFrictionFactor)
        {
            position = pos;
            maximumAirSpeed = maxAirSpeed;
            airControl = airControlFactor;
            maximumRunSpeed = maxRunSpeed;
            friction = frictionCoefficient;
            extraSlowDownFriction = extraSlowDown;
            maximumCorrectiveImpulse = maxCorrectiveImpulse;
            positionCorrectionMargin = positionCorrectionSlop;
            positionCorrectionFactor = positionCorrectionStrength;
            surfaceFrictionMultiplier = surfaceFrictionFactor;
        }

        /// <summary>
        /// Moves the character controller.
        /// </summary>
        /// <param name="dt">Time since last frame in seconds.</param>
        /// <param name="timeScale">Run speed factor.</param>
        /// <param name="timeSinceLastFrame">Time since last frame in real seconds.</param>
        public override void updateAtEndOfUpdate(float dt, float timeScale, float timeSinceLastFrame)
        {
            if (!isActive)
                return;
            //Don't let the feet detector move around by itself.
            currentState.feetShape.linearVelocity = Vector3.Zero;
            currentState.feetShape.angularVelocity = Vector3.Zero;
            rotationMatrix = Matrix.CreateFromQuaternion(orientationQuaternion);
            Vector3 bodyOffset = Vector3.Transform(currentState.shapeOffset, rotationMatrix);
            Vector3 feetOffset = Vector3.Transform(currentState.feetOffset, rotationMatrix);
            float impulse;
            float deltaV;
            Vector3 Ra, Rb;
            //The direction to push the character away from its supports.
            Vector3 norm = Vector3.Transform(Vector3.Up, rotationMatrix);
            float impulseTransform;

            supportLocations.Clear();
            supportNormals.Clear();
            supportTois.Clear();
            supportEntities.Clear();
            supportDepths.Clear();
            //Assembles some useful data about the supports below the character.
            getSupportRayHits(ref rotationMatrix, supportLocations, supportNormals, supportTois, supportDepths, supportEntities);
            //Keep the character 'hovering' above the ground.
            //Assumes that 'above' is just the positive y axis.
            for (int k = 0; k < supportTois.Count; k++)
            {

                //Calculate the upward force required to keep the player from getting pushed down anymore.
                Ra = supportLocations[k] - position;
                Rb = supportLocations[k] - currentState.shape.centerPosition;

                impulseTransform = 1 / currentState.shape.mass + 1 / supportEntities[k].mass +
                            Vector3.Dot(Vector3.Cross(Vector3.Transform((Vector3.Cross(Rb, supportNormals[k])), supportEntities[k].parent.inertiaTensorInverse), Rb), norm);
                deltaV = Vector3.Dot(Vector3.Cross(currentState.shape.angularVelocity, Rb) - currentState.shape.linearVelocity, norm);
                deltaV = Math.Max(deltaV, 0); //The character should be allowed to move up, away from the support.
                impulse = (deltaV) / (impulseTransform * supportTois.Count);


                if (supportEntities[k].isPhysicallySimulated)
                    supportEntities[k].applyImpulse(supportLocations[k], norm * (-impulse));
                currentState.shape.applyLinearImpulse(norm * impulse);
                //Applies a separate push to correct any position error.  This keeps the player floating at the top of the support ray rather than halfway down it.
                currentState.shape.correctiveLinearVelocity += (supportDepths[k] / supportTois.Count) * norm;
              

            }
            //Handle friction and horizontal motion.
            if (supportEntities.Count > 0)
            {
                Vector3 avgVelDifference = Vector3.Zero;
                foreach (Entity e in supportEntities)
                {
                    avgVelDifference += e.linearVelocity;
                }
                avgVelDifference /= supportEntities.Count;
                avgVelDifference -= currentState.shape.linearVelocity;
                if (avgVelDifference.LengthSquared() > Toolbox.bigEpsilon)
                {
                    float avgVelDirMagnitude = avgVelDifference.Length();
                    Vector3 avgVelDir = avgVelDifference / avgVelDirMagnitude;

                    for (int k = 0; k < supportEntities.Count; k++)
                    {

                        Ra = supportLocations[k] - position;
                        Rb = supportLocations[k] - supportEntities[k].centerPosition;
                        impulseTransform = 1 / currentState.shape.mass + 1 / supportEntities[k].mass +
                                   Vector3.Dot(Vector3.Cross(Vector3.Transform((Vector3.Cross(Rb, avgVelDir)), supportEntities[k].inertiaTensorInverse), Rb), avgVelDir);
                        impulse = avgVelDirMagnitude / impulseTransform;
                        if(!wasMoved) //Apply an extra friction if the character is not trying to move.
                            impulse *= (friction + extraSlowDownFriction + supportEntities[k].dynamicFriction * surfaceFrictionMultiplier) * .5f * dt;
                        else
                            impulse *= (friction + supportEntities[k].dynamicFriction * surfaceFrictionMultiplier) * .5f * dt;

                        supportEntities[k].applyImpulse(supportLocations[k], -impulse * avgVelDir);
                        currentState.shape.applyLinearImpulse(impulse * avgVelDir);

 

                    }

                }
            }


            //Update the positions
            position = currentState.shape.centerPosition - bodyOffset;
            currentState.feetShape.moveTo(position + feetOffset);
            currentState.feetShape.orientationQuaternion = orientationQuaternion;

            wasMoved = false;


        }
        /// <summary>
        /// Pushes the character in a given direction.
        /// The force is scaled down according to the airControl factor if the character is not touching anything.
        /// </summary>
        /// <param name="direction">Direction of the impulse.  Magnitude of direction corresponds to the magnitude of the impulse.</param>
        /// <param name="pushBack">Whether or not to push back in the opposite direction on the support entities.</param>
        /// <param name="useFriction">Whether or not to use the friction coefficient combined with the surface's friction.
        /// If pushback is disabled, this has no effect.</param>
        public void applyMovementImpulse(Vector3 direction, bool pushBack, bool useFriction)
        {
            float dot;
            float directionMagnitude = direction.Length();
            float maxSpeed;
            float factor;
            //Setup some initial information.
            Vector3 dir = direction / directionMagnitude;
            if (supportEntities.Count > 0)
            {
                Vector3 avgVel = Vector3.Zero;
                foreach (Entity e in supportEntities)
                {
                    avgVel += e.linearVelocity;
                }
                avgVel /= supportEntities.Count;
                maxSpeed = maximumRunSpeed + Vector3.Dot(avgVel, dir);
                factor = 1;
            }
            else
            {
                if (airControl == 0)
                    return;
                factor = airControl;
                maxSpeed = maximumAirSpeed;
            }

            //If there's supports and opposite forces are supposed to be applied to them..
            if (pushBack && supportEntities.Count > 0)
            {

                direction /= supportEntities.Count;
                Vector3 Ra, Rb;
                float impulseTransform;
                float impulse;
                for (int k = 0; k < supportEntities.Count; k++)
                {

                    Ra = supportLocations[k] - position;
                    Rb = supportLocations[k] - supportEntities[k].centerPosition;
                    impulseTransform = 1 / currentState.shape.mass + 1 / supportEntities[k].mass +
                               Vector3.Dot(Vector3.Cross(Vector3.Transform((Vector3.Cross(Rb, direction)), supportEntities[k].inertiaTensorInverse), Rb), dir);
                    impulse = directionMagnitude / impulseTransform;

                    if (useFriction)
                    {
                        impulse *= (friction + supportEntities[k].dynamicFriction) * .5f;
                    }

                    dot = Math.Max(Vector3.Dot(currentState.shape.linearVelocity, dir), 0);
                    if (dot < maxSpeed)
                    {
                        impulse = Math.Min(impulse, (maxSpeed - dot) * currentState.shape.mass );
                        supportEntities[k].applyImpulse(supportLocations[k], -impulse * dir);
                        currentState.shape.applyLinearImpulse(impulse * dir);

                    }
                }


            }
            else
            {//If there are no supports, or pushback is disabled, just 'dumbly' push the character forward.
                dot = Math.Max(Vector3.Dot(currentState.shape.linearVelocity, dir), 0);
                if (dot < maxSpeed)
                    currentState.shape.applyLinearImpulse(Math.Min(directionMagnitude, (maxSpeed - dot) * currentState.shape.mass * factor) * dir);
            }
            wasMoved = true;

        }

        /// <summary>
        /// Applies the impulse to the character controller shape, changing its momentum instantly.
        /// </summary>
        /// <param name="pos">Origin of the impulse.</param>
        /// <param name="direction">Direction of the impulse.  Magnitude of direction corresponds to the magnitude of the impulse.</param>
        public void applyImpulse(Vector3 pos, Vector3 direction)
        {
            currentState.shape.applyImpulse(pos, direction);
        }

        /// <summary>
        /// Jumps the character with a velocity in the direction given with a magnitude equal to the direction's length,
        /// provided that its feet are in contact with some object.
        /// </summary>
        /// <param name="direction">Direction to jump.</param>
        public void jump(Vector3 direction)
        {
            
            if (supportEntities.Count > 0)
            {
                direction /= supportEntities.Count;
                Vector3 Ra, Rb;
                float impulseTransform;
                Vector3 toApply;
                bool usePhysicalVersion;
                for (int k = 0; k < supportEntities.Count; k++)
                {

                    usePhysicalVersion = true;
                    if (supportEntities[k].isPhysicallySimulated)
                    {
                        foreach (Controller c in supportEntities[k].controllers)
                        {
                            if (c.colliderA != currentState.feetShape && c.colliderA != currentState.shape &&
                                c.colliderB != currentState.feetShape && c.colliderB != currentState.shape)
                                usePhysicalVersion = false;
                        }
                    }
                    else
                        usePhysicalVersion = false;
                    if (usePhysicalVersion)
                    {
                        //If the entity is 'floating' by itself, it can be pushed down in a physically correct manner, so that not much jump boost is gained.
                        Ra = supportLocations[k] - position;
                        Rb = supportLocations[k] - supportEntities[k].centerPosition;
                        impulseTransform = 1 / currentState.shape.mass + 1 / supportEntities[k].mass +
                                   Vector3.Dot(Vector3.Cross(Vector3.Transform((Vector3.Cross(Rb, direction)), supportEntities[k].inertiaTensorInverse), Rb), Vector3.Normalize(direction));
                        toApply = direction / impulseTransform;

                        currentState.shape.applyLinearImpulse(toApply);
                        supportEntities[k].applyImpulse(supportLocations[k], -toApply);
                    }
                    else
                    {//If the entity is touching something, assume it can transfer the jump force and it acts as an immobile object so that the full jump boost is gained.
                        currentState.shape.applyLinearImpulse(direction * currentState.shape.mass);
                        supportEntities[k].applyImpulse(supportLocations[k], direction * (-currentState.shape.mass));
                    }

                }
                
                
            }
        }

        /// <summary>
        /// Collects locations, normals, and hit distances from the rays supporting the character controller.
        /// </summary>
        /// <param name="rotationMatrix">Matrix representing the orientation of the character controller.</param>
        /// <param name="locs">Locations of intersection.</param>
        /// <param name="normals">Normals of ray intersection.</param>
        /// <param name="tois">Distances along the rays until impact in units of the ray direction's length.</param>
        /// <param name="depths">Amount of penetration detected by the rays.</param>
        /// <param name="supports">List of supporting entities that intersected rays.</param>
        public void getSupportRayHits(ref Matrix rotationMatrix, List<Vector3> locs, List<Vector3> normals, List<float> tois, List<float> depths, List<Entity> supports)
        {
            Entity supportBodyCandidate;
            Vector3 hitLoc;
            Vector3 hitNorm;
            float toi;
            Vector3 o, d;
            foreach (Ray r in currentState.rays)
            {
                foreach (Controller c in currentState.feetShape.controllers)
                {
                    if (c.colliderA != currentState.feetShape && c.colliderA != currentState.shape && c.colliderB != currentState.shape)
                        supportBodyCandidate = c.colliderA;
                    else if (c.colliderB != currentState.feetShape && c.colliderA != currentState.shape && c.colliderB != currentState.shape)
                        supportBodyCandidate = c.colliderB;
                    else
                        continue;
                    //Be sure that the target is the highest ranked compound body if the target is a child of a compound...
                    //If a child is targeted, the calculations will be ineffective.
                    CompoundBody parent = Toolbox.getHighestRankCompoundBody(supportBodyCandidate);
                    if (parent != null)
                        supportBodyCandidate = parent;
                    o = position + Vector3.Transform(r.Position, rotationMatrix);
                    d = Vector3.Transform(r.Direction, rotationMatrix);
                    if (Toolbox.rayCast(o, d, 1, supportBodyCandidate, currentState.margin, out hitLoc, out hitNorm, out toi))
                    {
                        if (toi > 0)
                        {
                            hitNorm.Normalize();
                            hitLoc += -hitNorm * currentState.margin;
                        }
                        supports.Add(supportBodyCandidate);
                        depths.Add(d.Length() + currentState.margin - Vector3.Distance(o, hitLoc));
                        locs.Add(hitLoc);
                        normals.Add(hitNorm);
                        tois.Add(toi);
                    }
                }
            }

        }

        /// <summary>
        /// Moves the character controller discontinuously along the given vector.
        /// </summary>
        /// <param name="v">Vector to move by.</param>
        public void move(Vector3 v)
        {
            currentState.shape.move(ref v);
            currentState.feetShape.move(ref v);
            position += v;
        }

        /// <summary>
        /// Moves the character so that its center is located at the given vector.
        /// </summary>
        /// <param name="v">Location to move to.</param>
        public void moveTo(Vector3 v)
        {
            Vector3 offset = v - position;
            move(offset);
        }

        /// <summary>
        /// Sets the current state to the given state.
        /// </summary>
        /// <param name="stateName">Name of the state to switch to.</param>
        public void setState(string stateName)
        {
            isActive = true;
            //Keep the velocities going smoothly and preserve momentum.
            CharacterControllerState nextState = states[stateName];
            if (currentState != null)
            {
                nextState.shape.linearVelocity = currentState.shape.linearVelocity;
                nextState.shape.angularVelocity = currentState.shape.angularVelocity;
                mySpace.remove(currentState.feetShape);
                mySpace.remove(currentState.shape);
            }
            else
            {
                nextState.shape.linearVelocity = Vector3.Zero;
                nextState.shape.angularVelocity = Vector3.Zero;
            }
            //Move the member entities around appropriately.
            Matrix rotMat = Matrix.CreateFromQuaternion(orientationQuaternion);
            nextState.shape.moveTo(position);
            nextState.feetShape.moveTo(position);
            nextState.shape.orientationQuaternion = orientationQuaternion;
            nextState.feetShape.orientationQuaternion = orientationQuaternion;
            nextState.shape.move(Vector3.Transform(nextState.shapeOffset, rotMat));
            nextState.feetShape.move(Vector3.Transform(nextState.feetOffset, rotMat));
            nextState.shape.isAlwaysActive = true;
            mySpace.add(nextState.shape);
            mySpace.add(nextState.feetShape);
            currentState = nextState;
        }

        /// <summary>
        /// Removes the controller's entities from the space, stopping its updates.
        /// To reactivate it, reset the state by calling setState.
        /// </summary>
        public void deactivate()
        {
            if (isActive)
            {
                isActive = false;
                mySpace.remove(currentState.feetShape);
                mySpace.remove(currentState.shape);
                currentState = null;
            }
        }


    }
}
