/*
   BEPUphysicsDemos Project

      Copyright (C) 2008 Bepu Entertainment LLC

      This demonstration software, excluding the BEPUphysics.dll and BEPUphysics.xml (which are handled independently by their accompanying license),
      is provided 'as-is', without any express or implied warranty.  In no event will the authors be held liable for any damages
      arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
    Ross Nordby
    norbo@bepu-games.com
    Harrison Nordby
    squashwell@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using System.Diagnostics;
using BEPUphysics.Entities;

namespace BEPUphysicsDemos
{
    /// <summary>
    /// Handles input and movement of a character in the game.
    /// Acts as the 'front end' for the bookkeeping and math of the dynamic character controller.
    /// </summary>
    public class DynamicCharacterInput
    {
        /// <summary>
        /// Physics representation of the character.
        /// </summary>
        public DynamicCharacterController characterController = new DynamicCharacterController();
        /// <summary>
        /// Current offset from the position of the character to the 'eyes.'
        /// </summary>
        public Vector3 cameraOffset;
        /// <summary>
        /// Offset from the position of the character to its 'eyes' while standing.
        /// </summary>
        public Vector3 standCameraOffset = new Vector3(0, 1.5f, 0);
        /// <summary>
        /// Offset from the position of the character to its 'eyes' while crouching.
        /// </summary>
        public Vector3 crouchCameraOffset = Vector3.Zero;
        /// <summary>
        /// Camera to use for input.
        /// </summary>
        public Camera camera;
        /// <summary>
        /// Owning space of the character.
        /// </summary>
        public Space space;
        /// <summary>
        /// Standing state of the character.
        /// </summary>
        public CharacterControllerState standState;
        /// <summary>
        /// Crouching state of the character.
        /// </summary>
        public CharacterControllerState crouchState;

        /// <summary>
        /// Whether or not to use the character controller's input.
        /// </summary>
        public bool isActive = false;

        //Movement forces
        float forwardForce = 20;
        float backForce = 15;
        float rightForce = 15;
        float leftForce = 15;
        float jumpImpulse = 7;
        
        //Amount of extra height to gain when performing a duck-jump.
        Vector3 jumpDuckHeightBoost = new Vector3(0,1,0);




        /// <summary>
        /// Constructs the character and internal physics character controller.
        /// </summary>
        /// <param name="owningSpace">Space to add the character to.</param>
        /// <param name="cameraToUse">Camera to attach to the character.</param>
        public DynamicCharacterInput(Space owningSpace, Camera cameraToUse)
        {
            //The support rays are rays which come out the bottom of the feet
            //and push the character away from the ground.  They also help generate the contact points for jumping and running.
            List<Ray> standSupportRays = new List<Ray>();
            standSupportRays.Add(new Ray(new Vector3(0, -2f, 0), new Vector3(0, -.5f, 0)));

            List<Ray> crouchSupportRays = new List<Ray>();
            crouchSupportRays.Add(new Ray(new Vector3(0, -2f, 0), new Vector3(0, -.5f, 0)));
 

            //Gives the space the character controller.
            owningSpace.add(characterController);
            //Sets up some of the available tuning variables for the character.
            characterController.setData(Vector3.Zero, 10, .4f, 10, 15, 10f, 10, 2, .02f, 1);
            characterController.deactivate();
            //Create the standing state (tall capsule)
            Entity standShape = new Capsule(Vector3.Zero, 3, 1, 5);
            //Create the crouching state (short capsule)
            Entity crouchShape = new Capsule(Vector3.Zero, 1.5f, 1, 5);
            //If you set a row in the local inertia tensor inverse to zero, it will not rotate around that local axis.  
            //For example, if row 1 was 0's, the object would not rotate around the local X axis.
            //For this character, the rotation will be completely disabled by setting the whole inertia tensor inverse to zero.
            //An axis can also be restricted by using the RotationalAxisConstraint.
            standShape.localSpaceInertiaTensorInverse = Toolbox.zeroMatrix;
            crouchShape.localSpaceInertiaTensorInverse = Toolbox.zeroMatrix;
            //Create the character states.
            standState = new CharacterControllerState(standShape, Vector3.Zero, standSupportRays, 0.7f);
            crouchState = new CharacterControllerState(crouchShape, new Vector3(0,-.75f,0), crouchSupportRays, 0.7f);
            //Add the states to the character controller.
            characterController.states.Add("stand", standState);
            characterController.states.Add("crouch", crouchState);
            cameraOffset = standCameraOffset;

            space = owningSpace;
            

            camera = cameraToUse;
        }

        /// <summary>
        /// Gives the character control over the camera and movement input.
        /// </summary>
        public void activate()
        {
            if (!isActive)
            {
                isActive = true;
                camera.useMovementControls = false;
                //Switches the character controller to the specified state.  You could have multiple states, such as crouch and prone.
                //Switching to another state preserves motion, so if you switch to prone while running in stand mode, the momentum of the run would carry through until lost through friction.
                characterController.setState("stand");
                //Put the controller where the camera is.
                characterController.moveTo(camera.position - cameraOffset);


            }

        }

        /// <summary>
        /// Returns input control to the camera.
        /// </summary>
        public void deactivate()
        {
            if (isActive)
            {
                isActive = false;
                camera.useMovementControls = true;
                characterController.deactivate();
            }
        }

       

        //Prevents spam-jumping; if the player jumps immediately after jumping, 
        //it is possible that they could still be in contact with the ground and thus get a double (or more) boost.
        float timeSinceLastJump = 0;
        /// <summary>
        /// Handles the input and movement of the character.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        public void update(float dt, KeyboardState keyboardInput, GamePadState gamePadInput)
        {
            if (isActive)
            {

                //Note that the character controller's update method is not called here; this is because it is handled within its owning space.
                //This method's job is simply to tell the character to move around based on the camera and input.

                //Puts the camera at eye level, using the character shape's position.
                camera.position = characterController.currentState.shape.centerPosition - Vector3.Transform(characterController.currentState.shapeOffset, characterController.currentState.shape.orientationMatrix) + cameraOffset;
                Vector3 totalMovement = Vector3.Zero;
                timeSinceLastJump += dt;
#if XBOX360
                Vector3 forward = camera.getForwardVector();
                forward.Y = 0;
                forward.Normalize();
                Vector3 right = camera.getRightVector();
                totalMovement += gamePadInput.ThumbSticks.Left.Y * forwardForce * forward;
                totalMovement += gamePadInput.ThumbSticks.Left.X * rightForce * right;
                characterController.applyMovementImpulse(totalMovement * dt, true, true);
                //Jumping
                if (gamePadInput.IsButtonDown(Buttons.LeftStick) && timeSinceLastJump > .3f)
                {
                    timeSinceLastJump = 0;
                    characterController.jump(Vector3.Up * jumpImpulse);
                }
                //Crouching and Standing
                if (gamePadInput.Buttons.RightStick == ButtonState.Pressed && characterController.currentState != crouchState)
                {
                    cameraOffset = crouchCameraOffset;
                    characterController.setState("crouch");
                    //If the character is in the air when it crouches, give it a 'jump duck' capability by moving it up a little.
                    if (characterController.supportEntities.Count == 0)
                        characterController.move(jumpDuckHeightBoost);
                }
                else if (gamePadInput.Buttons.RightStick == ButtonState.Released && characterController.currentState == crouchState)
                {

                    cameraOffset = standCameraOffset;
                    characterController.setState("stand");
                    //To counteract the jump duck air boost (and to keep players from ducking their way into the sky), 
                    //remove the jump duck boost if the character is still in the air when it stands.
                    if (characterController.supportEntities.Count == 0)
                        characterController.move(-jumpDuckHeightBoost);

                }
#else


                //Collect the movement impulses.

                Vector3 movementDir;

                if (keyboardInput.IsKeyDown(Keys.E))
                {
                    movementDir = camera.getForwardVector();
                    movementDir.Y = 0;
                    movementDir.Normalize();
                    totalMovement+= movementDir * (forwardForce);
                }
                if (keyboardInput.IsKeyDown(Keys.D))
                {
                    movementDir = camera.getBackVector();
                    movementDir.Y = 0;
                    movementDir.Normalize();
                    totalMovement += movementDir * (backForce);
                }
                if (keyboardInput.IsKeyDown(Keys.S))
                {
                    movementDir = camera.getLeftVector();
                    movementDir.Y = 0;
                    movementDir.Normalize();
                    totalMovement += movementDir * (leftForce);
                }
                if (keyboardInput.IsKeyDown(Keys.F))
                {
                    movementDir = camera.getRightVector();
                    movementDir.Y = 0;
                    movementDir.Normalize();
                    totalMovement += movementDir * (rightForce);
                }
                float impulseMagnitude = totalMovement.Length();
                if (impulseMagnitude > 0)
                {
                    if (impulseMagnitude > forwardForce)
                    {//Restrict the character from getting more impulse from strafing and moving forward than simply moving forward.
                        totalMovement = totalMovement * (forwardForce / impulseMagnitude); //Clamps the magnitude to a maximum of the forward force.
                    }
                    //Apply movement
                    characterController.applyMovementImpulse(totalMovement * dt, true, true);
                }

                //Jumping
                if (keyboardInput.IsKeyDown(Keys.A) && timeSinceLastJump > .3f)
                {
                    timeSinceLastJump = 0;
                    characterController.jump(Vector3.Up * jumpImpulse);
                }

                //Note that for a fully robust character, depending on the game, it might be wise to verify whether or not
                //a character's new state could fit before switching to it (could use raycasts or swept collision detection).
                //A similar check would be useful for verifying jump-duck height resets.

                //Crouching and Standing
                if (keyboardInput.IsKeyDown(Keys.Z) && characterController.currentState != crouchState)
                {
                    cameraOffset = crouchCameraOffset;
                    characterController.setState("crouch");
                    //If the character is in the air when it crouches, give it a 'jump duck' capability by moving it up a little.
                    if (characterController.supportEntities.Count == 0)
                        characterController.move(jumpDuckHeightBoost);
                }
                else if (!keyboardInput.IsKeyDown(Keys.Z) && characterController.currentState == crouchState)
                {

                    cameraOffset = standCameraOffset;
                    characterController.setState("stand");
                    //To counteract the jump duck air boost (and to keep players from ducking their way into the sky), 
                    //remove the jump duck boost if the character is still in the air when it stands.
                    if (characterController.supportEntities.Count == 0)
                        characterController.move(-jumpDuckHeightBoost);

                }
#endif
            }
        }
    }
}
