using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;

namespace BEPUphysicsDemos
{
    public class KinematicCharacterController:Updateable
    {
        /// <summary>
        /// Entity used as the character's body.
        /// Should be physically simulated so that it can collect forces and do the math-side heavy lifting.
        /// Needs to have isDetector = true so that contacts created by the controller are not solved internally.
        /// </summary>
        Entity bodyShape;

        /// <summary>
        /// Produces a bounding volume which encloses the character and all of its possible movement.
        /// This both detects nearby objects and notifies proximity based systems like terrain and static triangle groups to create entities.
        /// </summary>
        Entity characterBounds;
        /// <summary>
        /// Position of the character.
        /// </summary>
        public Vector3 position;
        /// <summary>
        /// Offset from the position of the character to the geometric center of the feet shape.
        /// </summary>
        Vector3 feetOffset;
        /// <summary>
        /// Shape of the feet to use in sweeping down to find support objects.
        /// </summary>
        Entity feetShape;
        /// <summary>
        /// Distance to sweep the feet shape downwards from its offset divided by two.
        /// The upper half is used to determine supports and keep the character a certain height upwards,
        /// while the lower half allows the entity to step downward.
        /// </summary>
        float maximumStepHeight;

        /// <summary>
        /// Movement direction and magnitude.
        /// </summary>
        Vector3 velocity;
        /// <summary>
        /// Maximum speed the character can achieve under its own power.
        /// </summary>
        float maximumVelocity;
        /// <summary>
        /// Rate of change of velocity while being ordered to change speeds.
        /// </summary>
        float acceleration;

        /// <summary>
        /// The entity hit first by the feet shape's convex cast.
        /// </summary>
        Entity supportingEntity;
        /// <summary>
        /// Surface normal at the point of contact between the characters feet and the ground.
        /// </summary>
        Vector3 supportNormal;
        /// <summary>
        /// Position of the contact between the character body and the supporting entity.
        /// </summary>
        Vector3 supportLocation;
        /// <summary>
        /// List of entities in the vicinity of the character.
        /// </summary>
        List<Entity> supportingCandidates = new List<Entity>();

        /// <summary>
        /// Creates a character controller.
        /// </summary>
        /// <param name="position">Initial position of the character's body.</param>
        /// <param name="height">Total height of the character's body/</param>
        /// <param name="width">Width of the character's body.</param>
        /// <param name="skinThickness">Margin around the body used to detect other objects prior to intersection with the main body.</param>
        /// <param name="maximumStepHeight">Maximum height the character controller can step up an obstacle or step down off of a ledge onto a lower surface.</param>
        /// <param name="maximumVelocity">Maximum speed the character can achieve under its own power.</param>
        /// <param name="acceleration">Rate of change of velocity while being ordered to change speeds.</param>
        /// <param name="mass">Mass of the character's body.</param>
        public KinematicCharacterController(Vector3 position, float height, float width, float skinThickness, float maximumStepHeight, float maximumVelocity, float acceleration, float mass)
        {
            this.position = position;
            this.maximumStepHeight = maximumStepHeight;
            this.maximumVelocity = maximumVelocity;
            //Create a body.  Decide capsule or cylinder standard.  Pass in height/width information.  Skin width.
            //Offsets and position are needed as well, but the feet shape itself can always be determined based on width (its going to be a disc-like shape).
            bodyShape = new Capsule(position, height - 2 * width, width, mass);
            bodyShape.collisionMargin = skinThickness;
            feetOffset = new Vector3(0, -height / 2 + maximumStepHeight, 0);
            feetShape = new Cylinder(position + feetOffset, 0, width + skinThickness);
            feetShape.collisionMargin = 0;
            characterBounds = new Box(position, width, height + 2 * maximumStepHeight, width, 1);
            characterBounds.isTangible = false;
            characterBounds.localSpaceInertiaTensorInverse = Toolbox.zeroMatrix;//avoids having to reorient the bounds every frame.
        }


        /// <summary>
        /// Updates the character.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        /// <param name="timeScale">Factor of the speed of the simulation.</param>
        /// <param name="timeSinceLastFrame">Time since last frame in real seconds.</param>
        public override void updateAtEndOfUpdate(float dt, float timeScale, float timeSinceLastFrame)
        {
            characterBounds.centerPosition = position;
            characterBounds.linearVelocity = velocity;
            characterBounds.findBoundingBox(dt);
            supportingCandidates.Clear();
            mySpace.broadPhase.getEntities(characterBounds.boundingBox, supportingCandidates);
            //Sweep forward as far as it can go based on its velocity and surroundings.
            advance(dt);
            float toi;
            if (findSupport(1, out supportingEntity, out supportLocation, out supportNormal, out toi))
            {//Put the character into contact with the ground.
                float heightIncreaseDesired = 1 - toi * 2 * maximumStepHeight;
                //try sweeping upwards.
                if (verifyHeadSafety(heightIncreaseDesired))
                {//Step up since it appears to be safe!
                    position += Vector3.Up * heightIncreaseDesired;
                }
            }
           
        }

        void advance(float dt)
        {
            Vector3 sweep = velocity * dt;
            /*float toi;
            float minimumToi = float.MaxValue;
            foreach (Entity e in supportingCandidates)
            {
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areSweptObjectsColliding(bodyShape, e, ref position, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,
                    0, e.collisionMargin, ref sweep, ref Toolbox.zeroVector, out toi))
                {
                    if (toi < minimumToi)
                        minimumToi = toi;
                }
            }*/
            position += sweep;// *minimumToi;
        }

        /// <summary>
        /// Check whether or not the body shape can move up the given distance without bumping its head.
        /// </summary>
        /// <param name="distance">Distance upwards to check.</param>
        /// <returns>Whether or not it is safe to move upwards.</returns>
        bool verifyHeadSafety(float distance)
        {
            Vector3 sweep = Vector3.Up * distance;
            float toi;
            foreach (Entity e in supportingCandidates)
            {
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areSweptObjectsColliding(bodyShape, e, ref position, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,
                    0, e.collisionMargin, ref sweep, ref Toolbox.zeroVector, out toi))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Finds the earliest supporting entity below the character, if any.
        /// </summary>
        /// <param name="candidates">A list of support entity candidates.</param>
        /// <param name="latestTimeAllowed">Latest time allowed for finding supports.</param>
        /// <param name="supportingEntity">The earliest supporting entity.</param>
        /// <param name="toi">The time of impact against the earliest supporting entity.</param>
        /// <returns>Whether or not any support was found.</returns>
        bool findSupport(float latestTimeAllowed, out Entity supportingEntity, out Vector3 location, out Vector3 normal, out float toi)
        {
            Vector3 feetPosition = position + feetOffset;
            Vector3 sweepDown = new Vector3(0, -2 * maximumStepHeight, 0);
            location = Toolbox.noVector;
            normal = Toolbox.noVector;
            supportingEntity = null;
            float minimumToi = float.MaxValue;
            foreach (Entity e in supportingCandidates)
            {
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areSweptObjectsColliding(feetShape, e, ref feetPosition, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,
                    0, e.collisionMargin, ref sweepDown, ref Toolbox.zeroVector, out location, out normal, out toi) && toi < minimumToi && toi > 0 && toi <= latestTimeAllowed) 
                {
                    minimumToi = toi;
                    supportingEntity = e;
                }
            }
            toi = minimumToi;
            return toi < float.MaxValue;
        }

        /// <summary>
        /// Accelerates the character towards the given direction.
        /// </summary>
        /// <param name="direction">Direction to accelerate.</param>
        public void accelerate(Vector3 direction)
        {
            direction.Normalize();
            velocity += direction * acceleration;
        }

        
    }
}
