using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using System.Diagnostics;
using BEPUphysics.Entities;

namespace BEPUphysicsDemos
{
    /// <summary>
    /// Velocity-level character
    /// </summary>
    public class KinematicCharacterController050:Updateable
    {
        /// <summary>
        /// Acts as the shape of the character in collisions.
        /// </summary>
        public Entity bodyShape;
        /// <summary>
        /// Collects information about the environment, including collision candidates and external forces.
        /// </summary>
        public Entity detector;
        /// <summary>
        /// Position of the character.
        /// </summary>
        public Vector3 position;
        /// <summary>
        /// Offset from the position of the character to the geometric center of the feet shape.
        /// </summary>
        Vector3 feetOffset;
        /// <summary>
        /// Shape of the feet to use in sweeping down to find support objects.
        /// </summary>
        Entity feetShape;

        #region Character Ability Variables
        /// <summary>
        /// Maximum speed that the character can achieve under its own power.
        /// </summary>
        public float maximumSpeed;
        /// <summary>
        /// Rate at which the character can change movement speeds.
        /// </summary>
        public float acceleration;
        /// <summary>
        /// Rate at which the character's velocity reduces in directions which the character is not trying to move.
        /// </summary>
        public float deceleration;
        /// <summary>
        /// Whether or not to modify the acceleration/deceleration of the character by the support's friction value.
        /// </summary>
        public bool useSupportFriction;
        /// <summary>
        /// Maximum height which the character is able to climb without jumping.
        /// </summary>
        public float maximumStepHeight;
        /// <summary>
        /// Initial boost speed in the up direction when jumping.
        /// </summary>
        public float jumpSpeed;
        /// <summary>
        /// Maximum speed the character can achieve under its own power while midair.
        /// </summary>
        public float maximumAirSpeed;
        /// <summary>
        /// Rate of change of velocity while airborne.
        /// </summary>
        public float airAcceleration;
        /// <summary>
        /// Angle between the up vector and the character's supporting terrain normal past which 
        /// gravity begins to affect the speed of the character.
        /// </summary>
        public float penaltySlope;
        /// <summary>
        /// Maximum angle between the normal of the source position of a step 
        /// (such as the ground when stepping up to the first stair in a staircase)
        /// and the up vector allowed.  If the angle is larger, the character will
        /// not be able to step to the target position.
        /// </summary>
        public float maximumStepSourceAngle;
        /// <summary>
        /// Maximum angle between the normal of the target position of a step
        /// (such as the top of a stair when stepping up from the ground)
        /// and the up vector allowed.  If the angle is larger, the character will
        /// not be able to step to the target position.
        /// </summary>
        public float maximumStepTargetAngle;
        /// <summary>
        /// Steepest slope which can be used as a support.
        /// </summary>
        public float maximumSlope;
        #endregion

        #region Movement Information
        /// <summary>
        /// Whether or not the character has a stable support.
        /// </summary>
        public bool isAirborne;
        /// <summary>
        /// Current linear velocity of the character.
        /// </summary>
        public Vector3 velocity;
        /// <summary>
        /// Direction in which the character attempts to walk.
        /// </summary>
        public Vector2 movementDirection;
        /// <summary>
        /// Current normal at the contact point between the character and the support.
        /// </summary>
        public Vector3 supportNormal;
        /// <summary>
        /// Entity currently supporting the character.
        /// </summary>
        public Entity supportEntity;
        /// <summary>
        /// Contact point between the character and supporting entity.
        /// </summary>
        public Vector3 supportLocation;
        /// <summary>
        /// List of entities that are in the vicinity of the character.
        /// </summary>
        public List<Entity> nearbyEntities = new List<Entity>();
        /// <summary>
        /// Angle between the normal of the support surface and the up vector.
        /// </summary>
        public float slope;
        #endregion

        /// <summary>
        /// Constructs a character which is moved by velocities instead of forces and impulses.
        /// </summary>
        /// <param name="position">Initial position of the character's body.</param>
        /// <param name="height">Total height of the character's body/</param>
        /// <param name="width">Width of the character's body.</param>
        /// <param name="skinThickness">Margin around the body used to detect other objects prior to intersection with the main body.</param>
        /// <param name="maxSpeed">Maximum speed that the character can achieve under its own power.</param>
        /// <param name="accel">Rate at which the character can change movement speeds.</param>
        /// <param name="decel">Rate at which the character slows down when not trying to move.</param>
        /// <param name="maxAirSpeed">Maximum speed the character can achieve under its own power while midair.</param>
        /// <param name="midAirAcceleration">Rate of change of velocity while airborne.</param>
        /// <param name="supportFriction">Whether or not to modify the acceleration/deceleration of the character by the support's friction value.</param>
        /// <param name="maximumStepHeight">Maximum height which the character is able to climb without jumping.</param>
        /// <param name="initialJumpSpeed">Initial boost speed in the up direction when jumping.</param>
        /// <param name="penaltyAngle">Angle between the up vector and the character's supporting terrain normal past which 
        /// gravity begins to affect the speed of the character.</param>
        /// <param name="maximumAngle">Steepest slope which can be used as a support.</param>
        /// <param name="maxStepSourceAngle">Maximum angle between the normal of the source position of a step 
        /// (such as the ground when stepping up to the first stair in a staircase)
        /// and the up vector allowed.  If the angle is larger, the character will
        /// not be able to step to the target position.</param>
        /// <param name="maxStepTargetAngle">Maximum angle between the normal of the target position of a step
        /// (such as the top of a stair when stepping up from the ground)
        /// and the up vector allowed.  If the angle is larger, the character will
        /// not be able to step to the target position.</param>
        public KinematicCharacterController050(Vector3 position, float height, float width, float skinThickness, float mass, float maxSpeed, float accel, float decel, float maxAirSpeed, float midAirAcceleration, bool supportFriction, float maximumStepHeight, float initialJumpSpeed, float penaltyAngle, float maximumAngle, float maxStepSourceAngle, float maxStepTargetAngle)
        {
            this.position = position;
            this.maximumStepHeight = maximumStepHeight;
            bodyShape = new Cylinder(position, height, width, mass);
            bodyShape.collisionMargin = skinThickness;
            feetOffset = new Vector3(0, -height / 2 + maximumStepHeight, 0);
            feetShape = new Cylinder(position + feetOffset, 0, width + skinThickness);
            feetShape.collisionMargin = 0;
            //Construct the detector entity.
            //For simplicity, it is an intangible box which encompasses the character body and a margin up and down representing the step height.
            detector = new Box(position, width + skinThickness, height + maximumStepHeight * 2, width + skinThickness, 1);
            detector.isAffectedByGravity = false;
            detector.isAlwaysActive = true;
            detector.isTangible = false;
            isAirborne = true;

            maximumSpeed = maxSpeed;
            acceleration = accel;
            deceleration = decel;
            maximumAirSpeed = maxAirSpeed;
            airAcceleration = midAirAcceleration;
            useSupportFriction = supportFriction;
            jumpSpeed = initialJumpSpeed;
            penaltySlope = penaltyAngle;
            maximumStepSourceAngle = maxStepSourceAngle;
            maximumStepTargetAngle = maxStepTargetAngle;
            maximumSlope = maximumAngle;
        }

        /// <summary>
        /// Sets the owning space of the updateable to the given space.
        /// Called automatically when Space.add(Updateable) is used.
        /// </summary>
        /// <param name="newSpace">The new owner of the updateable.</param>
        public override void addToSpace(Space newSpace)
        {
            base.addToSpace(newSpace);
            newSpace.add(detector);
        }

        /// <summary>
        /// Prepares the updateable for removal from the space.
        /// Called automatically when Space.remove(Updateable) is used.
        /// </summary>
        public override void removeFromSpace()
        {
            mySpace.remove(detector);
            base.removeFromSpace();
        }
        /// <summary>
        /// Updates the character.  Called automatically by the owning Space.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        /// <param name="timeScale">Factor of the speed of the simulation.</param>
        /// <param name="timeSinceLastFrame">Time since last frame in real seconds.</param>
        public override void updateAtEndOfUpdate(float dt, float timeScale, float timeSinceLastFrame)
        {
            nearbyEntities.Clear();
            foreach (Controller c in detector.controllers)
            {
                if (c.colliderA == detector)
                    nearbyEntities.Add(c.colliderB);
                else
                    nearbyEntities.Add(c.colliderA);
            }

            //supportNormal = Vector3.Up;
            velocityUpdate(dt);
            #region Position Update

            //Prepare to update the position and execute sweep tests

            stepUp();
            float toi;
            Vector3 originalLocation = bodyShape.centerPosition;
            if (!isAirborne)
            {
                //Sweep along the velocity.
                Entity hitEntity;
                Vector3 hitLocation;
                Vector3 hitNormal;
                if (findEarliestSweptIntersection(velocity * dt, out hitEntity, out hitLocation, out hitNormal, out toi))
                {
                    float velocityMagnitude = velocity.Length();
                    Vector3 velocityDirection;
                    if (velocityMagnitude != 0)
                        velocityDirection = velocity / velocityMagnitude;
                    else
                        velocityDirection = Vector3.Zero;
                    if (toi > 0)
                    {
                        Vector3 horizontalSweptLocation = originalLocation + velocityDirection * (velocityMagnitude * dt * toi);
                        bodyShape.moveTo(horizontalSweptLocation);
                    }
                    else
                        bodyShape.move(velocity * dt);
                }
                else
                {
                    //If the character can move unimpeded, just push it forward using its velocity.
                    bodyShape.move(velocity * dt);
                }




            }
            else
            {

                //Flying through the air, check to make sure it's not flying into something.
                /*if (findEarliestSweptIntersection(velocity * dt, out supportEntity, out supportLocation, out supportNormal, out toi))
                {

                    float velocityMagnitude = velocity.Length();
                    Vector3 velocityDirection = velocity / velocityMagnitude;
                    bodyShape.move(velocityDirection * (velocityMagnitude * dt * toi - bodyShape.collisionMargin));
                    if(supportNormal != Vector3.Zero)
                        supportNormal.Normalize();
                    else
                        bodyShape.move(-velocity * dt); //Get the body out of a strange rut.
                    if (Vector3.Dot(supportNormal, Vector3.Up) < 0)
                        supportNormal *= -1;
                    slope = (float)Math.Acos(Vector3.Dot(supportNormal, Vector3.Up));
                    if (Vector3.Dot(velocity, Vector3.Up) < 0 && slope < maximumSlope)
                    {//It's landed on an appropriate support.
                        isAirborne = false;
                    }
                    else
                    {
                        //Stop velocity that would cause interpenetration with the object.

                        velocity -= Vector3.Dot(supportNormal, velocity) * supportNormal;


                        supportEntity = null;
                        supportLocation = Toolbox.noVector;
                        supportNormal = Toolbox.noVector;
                    }

                }
                else
                    bodyShape.move(velocity * dt);*/

                if (findSupport(.5f, out supportEntity, out supportLocation, out supportNormal, out toi))
                {
                    if (supportNormal != Vector3.Zero)
                        supportNormal.Normalize();
                    //else
                    //    body.move(-velocity * dt); //Get the body out of a strange rut.
                    if (Vector3.Dot(supportNormal, Vector3.Up) < 0)
                        supportNormal *= -1;
                    slope = (float)Math.Acos(Vector3.Dot(supportNormal, Vector3.Up));
                    if (slope < maximumSlope)
                    {
                        isAirborne = false;
                        bodyShape.move(Vector3.Down * ((toi - .5f) * 2 * maximumStepHeight - 0));


                    }
                    else
                    {
                        //velocity += Vector3.Dot(velocity, supportNormal) * velocity;
                    }
                }
                else
                    bodyShape.move(velocity * dt);


            }
            #endregion
            Debug.WriteLine("___");
            Debug.WriteLine(bodyShape.centerPosition);
            Debug.WriteLine(velocity);

            if(supportNormal != Vector3.Zero)
                supportNormal.Normalize();
            //Align the normal.
            if (Vector3.Dot(Vector3.Up, supportNormal) < 0)
                supportNormal *= -1;
            //Update the detector state
            detector.moveTo(bodyShape.centerPosition);
            detector.linearVelocity = velocity;
            detector.angularVelocity = Vector3.Zero;
        }

        void velocityUpdate(float dt)
        {
            if (!isAirborne)
            {
                Vector3 supportVelocity = Toolbox.getVelocityOfPoint(supportLocation, supportEntity);
                //Identify the horizontal rotation axis and angle to compute the surface's rotation matrix.
                slope = (float)Math.Acos(Vector3.Dot(supportNormal, Vector3.Up));
                if (slope > maximumSlope)
                    isAirborne = true;
                else
                {
                    Matrix surfaceRotation;
                    Vector3 sideVector = Vector3.Cross(Vector3.Up, supportNormal);
                    if (sideVector == Vector3.Zero)
                        surfaceRotation = Matrix.Identity;
                    else
                    {
                        sideVector.Normalize();
                        surfaceRotation = Matrix.CreateFromAxisAngle(sideVector, slope);
                    }
                    //Find the movement direction on the surface of the support.
                    Vector3 movementDir = Vector3.Transform(new Vector3(movementDirection.X, 0, movementDirection.Y), surfaceRotation);
                    //Clamp movement velocity
                    float velocityMagnitude = Vector3.Dot(velocity, movementDir) - Vector3.Dot(supportVelocity, movementDir);
                    float nextVelocityMagnitude;
                    if (useSupportFriction)
                        nextVelocityMagnitude = Math.Min(velocityMagnitude + acceleration * dt * supportEntity.dynamicFriction, maximumSpeed);
                    else
                        nextVelocityMagnitude = Math.Min(velocityMagnitude + acceleration * dt, maximumSpeed);
                    float accelerationToUse = nextVelocityMagnitude - velocityMagnitude;
                    velocity += movementDir * accelerationToUse;
                    //Remove any motion parallel to the normal.
                    velocity -= (Vector3.Dot(velocity, supportNormal) * supportNormal - Vector3.Dot(supportVelocity, supportNormal) * supportNormal);
                    //Attempt to slow down any other motion that is not along the movement direction.
                    Vector3 otherMotion = velocity - movementDir * nextVelocityMagnitude - Vector3.Dot(supportVelocity, movementDir) * movementDir;
                    float otherMotionMagnitude = otherMotion.Length();
                    float newOtherMotionMagnitude = Math.Max(otherMotionMagnitude - deceleration * dt, 0);
                    float otherMotionDeceleration = newOtherMotionMagnitude - otherMotionMagnitude;
                    Vector3 otherMotionDirection;
                    if (otherMotionMagnitude == 0)
                        otherMotionDirection = Vector3.Zero;
                    else
                        otherMotionDirection = otherMotion / otherMotionMagnitude;

                    velocity += otherMotionDirection * otherMotionDeceleration;
                    //Check to see if the slope is steep enough to start applying gravity.
                    if (slope > penaltySlope)
                    {
                        //Find the amount of gravity to apply by finding the declining direction and then projecting gravity onto it.
                        Vector3 downSlope = Vector3.Cross(sideVector, supportNormal);
                        Vector3 gravityComponent = Vector3.Dot(downSlope, (new Vector3(0, -acceleration, 0) + mySpace.simulationSettings.gravity) * dt) * downSlope;
                        velocity += gravityComponent;
                    }
                }
            }
            else
            {
                //Floating mid-air, so apply air acceleration and clamp velocity
                Vector3 movementDir = new Vector3(movementDirection.X, 0, movementDirection.Y);
                float velocityMagnitude = Vector3.Dot(velocity, movementDir);
                float nextVelocityMagnitude = Math.Min(velocityMagnitude + airAcceleration * dt, maximumAirSpeed);
                float accelerationToUse = Math.Max(nextVelocityMagnitude - velocityMagnitude, 0);
                velocity += movementDir * accelerationToUse;
                //Apply gravity
                velocity += mySpace.simulationSettings.gravity * dt;
            }
        }

        void stepUp()
        {
            //Do the step process, if applicable.
            float toi;
            if (findSupport(1, out supportEntity, out supportLocation, out supportNormal, out toi))
            {
                if (!isAirborne || toi <= .5f)
                {
                    if (supportNormal != Vector3.Zero)
                        supportNormal.Normalize();
                    //else
                    //    body.move(-velocity * dt); //Get the body out of a strange rut.
                    if (Vector3.Dot(supportNormal, Vector3.Up) < 0)
                        supportNormal *= -1;
                    slope = (float)Math.Acos(Vector3.Dot(supportNormal, Vector3.Up));
                    if (slope < maximumSlope)
                    {
                        float stepAmount = ((toi - .5f) * 2 * maximumStepHeight);
                        if (toi > .5f || verifyHeadSafety(-stepAmount))
                            bodyShape.move(Vector3.Down * stepAmount);


                    }
                    else
                    {
                        isAirborne = true;
                        supportEntity = null;
                        supportLocation = Toolbox.noVector;
                        supportNormal = Toolbox.noVector;
                    }
                }
            }
            else
            {
                isAirborne = true;
                supportEntity = null;
                supportLocation = Toolbox.noVector;
                supportNormal = Toolbox.noVector;
            }
        }

        /// <summary>
        /// Check whether or not the body shape can move up the given distance without bumping its head.
        /// </summary>
        /// <param name="distance">Distance upwards to check.</param>
        /// <returns>Whether or not it is safe to move upwards.</returns>
        bool verifyHeadSafety(float distance)
        {
            Vector3 sweep = Vector3.Up * distance;
            float toi;
            Vector3 bodyPosition = bodyShape.centerPosition;
            foreach (Entity e in nearbyEntities)
            {
                if (e == supportEntity)
                    continue; //Supporting entities should be ignored as bonk-candidates, as this means that the body is stuffed into what it's going to be standing on.
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areSweptObjectsColliding(bodyShape, e, ref bodyPosition, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,
                    0, 0, ref sweep, ref Toolbox.zeroVector, out toi))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Finds the earliest supporting entity below the character, if any.
        /// </summary>
        /// <param name="candidates">A list of support entity candidates.</param>
        /// <param name="latestTimeAllowed">Latest time allowed for finding supports.</param>
        /// <param name="supportingEntity">The earliest supporting entity.</param>
        /// <param name="toi">The time of impact against the earliest supporting entity.</param>
        /// <returns>Whether or not any support was found.</returns>
        bool findSupport(float latestTimeAllowed, out Entity supportingEntity, out Vector3 location, out Vector3 normal, out float toi)
        {
            Vector3 feetPosition = bodyShape.centerPosition + feetOffset;
            Vector3 sweepDown = new Vector3(0, -2 * maximumStepHeight, 0);
            location = Toolbox.noVector;
            normal = Toolbox.noVector;
            Vector3 locationCandidate;
            Vector3 normalCandidate;
            supportingEntity = null;
            float minimumToi = float.MaxValue;
            foreach (Entity e in nearbyEntities)
            {
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areSweptObjectsColliding(feetShape, e, ref feetPosition, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,
                    0, e.collisionMargin, ref sweepDown, ref Toolbox.zeroVector, out locationCandidate, out normalCandidate, out toi) && toi < minimumToi && toi > 0 && toi <= latestTimeAllowed)
                {
                    location = locationCandidate;
                    normal = normalCandidate;
                    minimumToi = toi;
                    supportingEntity = e;
                }
            }
            toi = minimumToi;
            return toi < float.MaxValue;
        }

        /// <summary>
        /// Identifies the earliest collision between the body and the environment.
        /// </summary>
        /// <param name="sweepDirection">Direction to move the body of the character along.</param>
        /// <param name="hitEntity">First entity hit by the swept collision test.</param>
        /// <param name="hitLocation">Location at which the body and the environment hit.</param>
        /// <param name="normal">Normal of the contact between the body and the hit entity.</param>
        /// <param name="toi">Time of impact along the velocity to the impact. </param>
        /// <returns>Whether or not the body impacted anything in its environment.</returns>
        private bool findEarliestSweptIntersection(Vector3 sweepDirection, out Entity hitEntity, out Vector3 hitLocation, out Vector3 normal, out float toi)
        {
            hitEntity = null;
            hitLocation = Toolbox.noVector;
            normal = Toolbox.noVector;
            toi = float.NegativeInfinity;
            float minimumToi = float.MaxValue;
            Vector3 hitLocCandidate, normalCandidate;
            float toiCandidate;

            foreach (Entity candidate in nearbyEntities)
            {
                Toolbox.areSweptObjectsColliding(bodyShape, candidate, 0,0, ref sweepDirection, ref Toolbox.zeroVector, out hitLocCandidate, out normalCandidate, out toiCandidate);
                if (toiCandidate < minimumToi)
                {
                    minimumToi = toiCandidate;
                    hitLocation = hitLocCandidate;
                    hitEntity = candidate;
                    normal = normalCandidate;
                    toi = toiCandidate;
                }
            }

            return toi != float.NegativeInfinity;
        }

        public void jump()
        {
            if (supportEntity != null)
            {
                velocity.Y += jumpSpeed;
                supportEntity = null;
                supportLocation = Toolbox.noVector;
                supportNormal = Toolbox.noVector;
                isAirborne = true;
            }
        }

    }
}
