/*
   BEPUphysicsDemos Project

      Copyright (C) 2008 Bepu Entertainment LLC

      This demonstration software, excluding the BEPUphysics.dll and BEPUphysics.xml (which are handled independently by their accompanying license),
      is provided 'as-is', without any express or implied warranty.  In no event will the authors be held liable for any damages
      arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
    Ross Nordby
    norbo@bepu-games.com
    Harrison Nordby
    squashwell@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using System.Diagnostics;
using Microsoft.Xna.Framework.Graphics;
using BEPUphysics.Entities;
using BEPUphysics.Constraints;
using BEPUphysics.Vehicle;

namespace BEPUphysicsDemos
{
    /// <summary>
    /// Handles input and movement of a vehicle in the game.
    /// Acts as the 'front end' for the bookkeeping and math of the vehicle within the physics engine.
    /// </summary>
    public class VehicleInput
    {
        /// <summary>
        /// Physics representation of the vehicle.
        /// </summary>
        public Vehicle vehicle;
        /// <summary>
        /// Current offset from the position of the vehicle to the 'eyes.'
        /// </summary>
        public Vector3 cameraOffset;
        /// <summary>
        /// Camera to use for input.
        /// </summary>
        public Camera camera;
        /// <summary>
        /// Owning space of the vehicle.
        /// </summary>
        public Space space;
        /// <summary>
        /// Draws the body and wheels.
        /// </summary>
        public EntityDrawer entityDrawer;
        /// <summary>
        /// Whether or not to use the character controller's input.
        /// </summary>
        public bool isActive = false;

        //Movement accelerations
        float forwardAcceleration = 7;
        float backwardsAcceleration = -4;

        /// <summary>
        /// Maximum turn angle of the wheels.
        /// </summary>
        public float maxTurnAngles = (float)Math.PI / 6;
        /// <summary>
        /// List of graphical representations of the wheels on the vehicle.
        /// </summary>
        public List<DisplayModel> wheelModels;
        /// <summary>
        /// Helps keep the vehicle upright in cases where it would otherwise do barrel rolls.
        /// </summary>
        UprightConstraint uprightConstraint;




        /// <summary>
        /// Constructs the front end and the internal physics representation of the vehicle.
        /// </summary>
        /// <param name="owningSpace">Space to add the vehicle to.</param>
        /// <param name="cameraToUse">Camera to attach to the vehicle.</param>
        public VehicleInput(Space owningSpace, Camera cameraToUse, EntityDrawer drawer, Model wheelModel, Texture2D wheelTexture)
        {
            CompoundBody body = new CompoundBody();
            body.addBody(new Box(new Vector3(0,0,0), 3, 1f, 5, 100));
            body.addBody(new Box(new Vector3(0,.75f,-.5f), 3, .5f, 3f, 10));
            body.moveTo(new Vector3(10000, 0, 0)); //At first, just keep it out of the way.
            vehicle = new Vehicle(body, 20, 10);
            //Increase the inertia parallel to the car to prevent rolling a little.
            Matrix inertiaTensorInverse = vehicle.body.localSpaceInertiaTensorInverse;
            inertiaTensorInverse.M31 *=  .5f;
            inertiaTensorInverse.M32 *= .5f;
            inertiaTensorInverse.M33 *= .5f;
            vehicle.body.localSpaceInertiaTensorInverse = inertiaTensorInverse;
            vehicle.body.collisionMargin = .1f;
            //The two sections below correspond to ray cast wheels and convex cast wheels; 
            //Uncomment the one you'd like to use- note that the defaults are a little different, so handling won't be the same between the two.
            #region RayCastWheels
            //Note that when ray cast wheels become completely 'submerged' and the raycast can't get a good normal,
            //the negative ray direction vector is used.  This can cause bad effects if the ray starts outside
            //of the vehicle; specifically, if the car is turned on its side, it may fail to get a normal
            //and push itself horizontally around in circles, which looks rather silly.
            //To prevent such behavior, ensure that the start of the ray is slightly within the vehicle.
            /*
            Quaternion initialWheelRotation = Quaternion.CreateFromAxisAngle(Vector3.Forward, (float)Math.PI / 2);
            //front left wheel:
            vehicle.addWheel(new RayCastWheel(new Vector3(-1.4f, -.6f, 2.14f), initialWheelRotation,
                .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .6f, .02f, .5f, 500, .8f));
            //front right wheel:
            vehicle.addWheel(new RayCastWheel(new Vector3(1.4f, -.6f, 2.14f), initialWheelRotation,
                .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .6f, .02f, .5f, 500, .8f));
            //back left wheel:
            vehicle.addWheel(new RayCastWheel(new Vector3(-1.4f, -.6f, -1.91f), initialWheelRotation,
                .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .6f, .02f, .5f, 500, .8f));
            //back right wheel:
            vehicle.addWheel(new RayCastWheel(new Vector3(1.4f, -.6f, -1.91f), initialWheelRotation,
                .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .6f, .02f, .5f, 500, .8f));
            */
            #endregion
            #region ConvexCastWheels

            //Convex cast wheels detect collisions in a smoother manner, preventing the wheel from 'falling into' a thin, deep crevice.
            //Similar recommendations apply in regards to keeping the starting position of the object within the car.  With convex cast wheels,
            //if your wheel is poking out to the sides and you drive near an obstacle/wall, the wheels will view that object as a support
            //which can cause interesting flips.
            //front left wheel:
            
            Quaternion initialWheelRotation = Quaternion.CreateFromAxisAngle(Vector3.Forward, (float)Math.PI / 2);
            Entity wheelShape = new Cylinder(Vector3.Zero, .2f, .375f);
            wheelShape.collisionMargin = 0;
            vehicle.addWheel(new ConvexCastWheel(new Vector3(-1.3f, -.1f, 2.14f), initialWheelRotation,
                wheelShape, .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .8f, .02f, .5f, 500, .8f));
            //front right wheel:
            wheelShape = new Cylinder(Vector3.Zero, .2f, .375f);
            wheelShape.collisionMargin = 0;
            vehicle.addWheel(new ConvexCastWheel(new Vector3(1.3f, -.1f, 2.14f), initialWheelRotation,
                wheelShape, .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .8f, .02f, .5f, 500, .8f));
            //back left wheel:
            wheelShape = new Cylinder(Vector3.Zero, .2f, .375f);
            wheelShape.collisionMargin = 0;
            vehicle.addWheel(new ConvexCastWheel(new Vector3(-1.3f, -.1f, -1.91f), initialWheelRotation,
                wheelShape, .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .8f, .02f, .5f, 500, .8f));
            //back right wheel:
            wheelShape = new Cylinder(Vector3.Zero, .2f, .375f);
            wheelShape.collisionMargin = 0;
            vehicle.addWheel(new ConvexCastWheel(new Vector3(1.3f, -.1f, -1.91f), initialWheelRotation,
                wheelShape, .375f, new Vector3(0, -1, 0), new Vector3(0, 0, 1), .6f, 42, .8f, .02f, .5f, 500, .8f));
            
            #endregion

            space = owningSpace;
            uprightConstraint = new UprightConstraint(body, Vector3.Up, (float)Math.PI / 8, (float)Math.PI / 2f, 2000);
            space.add(uprightConstraint);

            space.add(vehicle);
            entityDrawer = drawer;
            entityDrawer.addEntity(vehicle.body);
            DisplayModel model;
            wheelModels = new List<DisplayModel>();
            for (int k = 0; k < 4; k++)
            {
                vehicle.wheels[k].detectorShape.tag = "noDisplayObject";
                model = new DisplayModel(wheelModel);
                entityDrawer.displayModels.Add(model);
                wheelModels.Add(model);
                model.setTexture(wheelTexture);
            }
           


            camera = cameraToUse;

        }

        /// <summary>
        /// Gives the vehicle control over the camera and movement input.
        /// </summary>
        public void activate()
        {
            if (!isActive)
            {
                isActive = true;
                camera.useMovementControls = false;
                //Put the vehicle where the camera is.
                vehicle.moveTo(camera.position - cameraOffset);
                vehicle.body.linearVelocity = Vector3.Zero;
                vehicle.body.angularVelocity = Vector3.Zero;
                vehicle.body.orientationQuaternion = Quaternion.Identity;
                camera.activateChaseCameraMode(vehicle.body, new Vector3(0, 1, 0), true, 10);
            }

        }

        /// <summary>
        /// Returns input control to the camera.
        /// </summary>
        public void deactivate()
        {
            if (isActive)
            {
                isActive = false;
                camera.useMovementControls = true;
                camera.deactivateChaseCameraMode();
            }
        }

        

        /// <summary>
        /// Handles the input and movement of the character.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        public void update(float dt, KeyboardState keyboardInput, GamePadState gamePadInput)
        {
            //Update the wheel's graphics.
            for (int k = 0; k < 4; k++)
            {
                wheelModels[k].worldMatrix = vehicle.wheels[k].worldMatrix;// Matrix.CreateFromAxisAngle(new Vector3(0, 0, 1), (float)Math.PI / 2) * vehicle.wheels[k].worldMatrix;
            }
            //Don't bother righting the vehicle if it's flipping through the air, or if all of its wheels are grounded anyway.
            int support = vehicle.supportedWheelCount;
            if (support == 1 || support == 2)
                uprightConstraint.isUpdating = true;
            else
                uprightConstraint.isUpdating = false;
            if (isActive)
            {
                //Reset wheel orientations.
                vehicle.wheels[0].wheelFacingAngle = 0;
                vehicle.wheels[1].wheelFacingAngle = 0;

                vehicle.accelerate(0);
#if XBOX360
                vehicle.accelerate(gamePadInput.Triggers.Right * forwardAcceleration + gamePadInput.Triggers.Left * backwardsAcceleration);
                if (gamePadInput.IsButtonDown(Buttons.LeftStick))
                    vehicle.brake(.5f);
                else
                    vehicle.releaseBrake();
                vehicle.wheels[0].wheelFacingAngle += (gamePadInput.ThumbSticks.Left.X * maxTurnAngles);
                vehicle.wheels[1].wheelFacingAngle += (gamePadInput.ThumbSticks.Left.X * maxTurnAngles);
#endif

                if (keyboardInput.IsKeyDown(Keys.E))
                {
                    vehicle.accelerate(forwardAcceleration);
                }
                if (keyboardInput.IsKeyDown(Keys.D))
                {
                    vehicle.accelerate(backwardsAcceleration);
                }
                if (keyboardInput.IsKeyDown(Keys.Space))
                {
                    vehicle.brake(.5f);
                }
                else
                    vehicle.releaseBrake();
                if (keyboardInput.IsKeyDown(Keys.S))
                {
                    vehicle.wheels[0].wheelFacingAngle -= maxTurnAngles;
                    vehicle.wheels[1].wheelFacingAngle -= maxTurnAngles;
                }
                if (keyboardInput.IsKeyDown(Keys.F))
                {
                    vehicle.wheels[0].wheelFacingAngle += maxTurnAngles;
                    vehicle.wheels[1].wheelFacingAngle += maxTurnAngles;
                }
                



             
            }
            else//Tell the vehicle to brake if it's not being controlled.
                vehicle.brake(.5f);
        }
    }
}
